/*
XalsaVol :  ALSA volume adjuster. It uses only X11 and ALSA.

Copyright (c) 2015 by kt8216@sdf.org

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Installation:
------------
Just compile this file and link it with both the ALSA and X11 libs.

  gcc -o xalsavol xalsavol.c -lX11 -lasound

There are no other files to install, just put the resulting binary
somewhere safe.

Usage:
------
There are no configuration files. All configuration is passed on the
command line. See the output of 

  xalsavol -g

for more details.
*/
#include <stdlib.h>
#include <alsa/asoundlib.h>
#include <syslog.h>
#include <string.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define CROSS_WIDTH 6
#define CROSS_LENGTH 20
#define MARGIN    10

static long int min_vol, max_vol, sound_volume;
static snd_mixer_t *handle;
static snd_mixer_selem_id_t *sid;
static snd_mixer_elem_t* elem;
static int sound_ok;

static int dbg;

static int min_y, max_y, delta_y;

static unsigned int
    bowidth  = 3,
    width    = 40,
    height   = 150;
static int x_pos, y_pos;

static unsigned int bgcolor, ficolor, licolor, bocolor, btcolor, bfcolor;
static Window window;
static Display *dpy;
static GC gc;

static XPoint cross_pt[13];
static XPoint tri_pt[4];
static XPoint rec_pt[5];

static int cros_pressed, btn_pressed;


static void sound_error_handler
        (const char *file, int line, const char *function, int err,
         const char *fmt, ...)
{
  va_list ap;
  char buf[2000];
  va_start(ap,fmt);
  memset(buf,0,sizeof(buf));
  vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap); 
  buf[sizeof(buf)-1]= 0;
  syslog(LOG_WARNING,"ALSA(%s@%s:%d) ERROR %d: %s",
               function, file, line, err, buf);
}

static void set_volume(long int V)
{
  if (sound_ok)
     snd_mixer_selem_set_playback_volume_all(elem, V);
}

static int get_volume()
{
  if (!elem) return 1;
  if (snd_mixer_selem_get_playback_volume(elem, 0, &sound_volume))
  {
    sound_volume= 0;
    return 1;
  }
  if (sound_volume<min_vol || sound_volume>max_vol) 
  {
    syslog(LOG_WARNING,"sound volume %d is out of limits [%d,%d]\n", 
                 sound_volume, min_vol, max_vol);
    sound_volume= 0;
    return 1;
  }
  return 0;
}

void init_sound(const char *devname, const char *elemname)
{
  int E;
  const char *ecode;

#define INTERR(V,U) do { ecode=V; if ((E=U)) goto fail; } while(0)
  if (!devname) devname= "default";
  if (!elemname) elemname= "Master";

  snd_lib_error= sound_error_handler;
  
  INTERR("mixer_open", snd_mixer_open(&handle, 0));
  INTERR("mixer_attach",snd_mixer_attach(handle, devname));
  INTERR("selem_register", snd_mixer_selem_register(handle, NULL, NULL));
  INTERR("mixer_load",snd_mixer_load(handle));
  INTERR("id_malloc", snd_mixer_selem_id_malloc(&sid));
  snd_mixer_selem_id_set_index(sid, 0);
  snd_mixer_selem_id_set_name(sid, elemname);
  elem= snd_mixer_find_selem(handle, sid);
  if (!elem)
  {
    syslog(LOG_WARNING,"can't find mixer element %s", elemname);
    goto fail;
  }
  INTERR("get_range",
      snd_mixer_selem_get_playback_volume_range(elem, &min_vol, &max_vol));

  if (max_vol<=min_vol)
  {
    syslog(LOG_WARNING,"bad minimum/maximum volumes min:%d max: %d\n",
                        min_vol, max_vol);
    goto fail;
  }

  INTERR("get_volume",get_volume());

  sound_ok= 1;
  return ;

fail:
  if (sid) snd_mixer_selem_id_free(sid);
  if (handle) snd_mixer_close(handle);
  handle= NULL; elem= NULL; sid= NULL;
  sound_ok= 0;
}



#if 0
void SetAlsaMasterVolume(long volume)
{

}
#endif

static void usage(char *prg)
{
  fprintf(stderr,
"usage: %s [options]\n"
"  options:\n"
"    -c <device_name>    (default)\n"
"    -e <element_name>   (Master)\n"
"    -d                  debug mode on\n"
"    -b <color>          background color\n"
"    -f <color>          fill color\n"
"    -l <color>          line color\n"
"    -r <color>          window border color\n"
"    -t <color>          indicator color\n"
"    -u <color>          indicator border color\n"
"    -z <integer>        (1) window border width\n"
"    -x <x_position>\n"
"    -y <y_position>\n"
"    -w <width>\n"
"    -h <height>\n"
"    -v                  version and copying info\n"
"Color options should be given as X11 names or as #RRggBB.\n" ,prg);
  exit(1);
}

static unsigned int color(const char* name, const char *defname)
{
  XColor scr, exa; 
  if (!XAllocNamedColor(dpy,
             DefaultColormap(dpy,DefaultScreen(dpy)),
                             name,&scr, &exa))
  {
    syslog(LOG_WARNING,"can't allocate color %s", name);
    if (!XAllocNamedColor(dpy, DefaultColormap(dpy,DefaultScreen(dpy)),
                          defname,&scr, &exa));
      syslog(LOG_WARNING,"can't allocate color %s", defname);
  }
  return scr.pixel;
}

static void cross(int x,int y, int L, int M, int *R)
{
  int U,Z,V;
  int A[2];
  int pX, pY;
  Z= M / 2.82843;
  U= L / 1.41421;

  if (U%2) U--;

  V= Z-U/2;

#define pt(D, dx,dy)  { pX= *(D)= pX+(dx); pY= *(D+1)= pY+ (dy); }

  pX= x; pY= y;

  pt(A, Z, -Z);   
  pt(R+0, -U/2,-U/2);
  pt(R+2, -V, V);
  pt(R+4, -V, -V);
  pt(R+6, -U, U);
  pt(R+8, V, V);
  pt(R+10, -V, V);
  pt(R+12, U, U);
  pt(R+14, V, -V);
  pt(R+16, V, V);
  pt(R+18, U, -U);
  pt(R+20, -V, -V);
  pt(R+22, V, -V);
#undef pt
}

static void init_rec_coords()
{
  int YC;
  double P;

  P= (double) (sound_volume) / (max_vol-min_vol+1);
  YC= max_y - delta_y*P ;
  rec_pt[0].x= width*0.1;   rec_pt[0].y= YC - 5;
  rec_pt[1].x= width*0.9;   rec_pt[1].y= rec_pt[0].y;
  rec_pt[2].x= rec_pt[1].x; rec_pt[2].y= YC+5;
  rec_pt[3].x= rec_pt[0].x; rec_pt[3].y= rec_pt[2].y;
  rec_pt[4]= rec_pt[0];
}

static void init_coords()
{
  static int cros_pos[24];
  int i;

  cross(width/2,CROSS_LENGTH/2+3, CROSS_WIDTH, CROSS_LENGTH, cros_pos);
  for(i=0;i<12;i++)
  {
    cross_pt[i].x= cros_pos[2*i];
    cross_pt[i].y= cros_pos[2*i+1];
  }
  cross_pt[12]= cross_pt[0];

  tri_pt[0].x= width *0.2;
  tri_pt[0].y= CROSS_LENGTH *1.4+3;
  tri_pt[1].x= width - width*0.2;
  tri_pt[1].y= tri_pt[0].y;
  tri_pt[2].x= tri_pt[1].x;
  tri_pt[2].y= height - CROSS_LENGTH*0.4;
  tri_pt[3]= tri_pt[0];


  min_y= tri_pt[0].y + MARGIN;
  max_y= tri_pt[2].y - MARGIN;
  delta_y= max_y - min_y +1;

  init_rec_coords();
}


static void draw_window(int x,int y,int w,int h)
{
  XRectangle rect;

  rect.x= x; rect.y= y; rect.width= w; rect.height= h;

  XSetForeground(dpy, gc, ficolor);
  XSetClipRectangles(dpy, gc, 0,0,&rect, 1,Unsorted);

  XClearArea(dpy, window, x,y,w,h, False);

  if (cros_pressed)
    XSetForeground(dpy, gc, bgcolor);
  else
    XSetForeground(dpy, gc, ficolor);

  XFillPolygon(dpy,window, gc, cross_pt, 12, Nonconvex, CoordModeOrigin);

  XSetForeground(dpy, gc, ficolor);
  XFillPolygon(dpy,window, gc, tri_pt, 3, Convex, CoordModeOrigin);

  XSetForeground(dpy, gc, licolor);
  XDrawLines(dpy, window, gc, cross_pt, 13, CoordModeOrigin);
  XDrawLines(dpy,window, gc, tri_pt, 4, CoordModeOrigin);

  if (btn_pressed)
    XSetForeground(dpy, gc, bgcolor);
  else
    XSetForeground(dpy, gc, btcolor);

  XFillPolygon(dpy,window, gc, rec_pt, 4, Convex, CoordModeOrigin);
  XSetForeground(dpy, gc, bfcolor);
  XDrawLines(dpy,window, gc, rec_pt, 5, CoordModeOrigin);
}

static void print_version()
{
  puts(
"XalsaVol-1.0\n\n"
"Copyright (c) 2015 by kt8216@sdf.org\n"
"\n"
"All rights reserved.\n"
"\n"
"Redistribution and use in source and binary forms, with or without\n"
"modification, are permitted provided that the following conditions\n"
"are met:\n"
"\n"
"1. Redistributions of source code must retain the above copyright notice,\n"
"   this list of conditions and the following disclaimer.\n"
"\n"
"2. Redistributions in binary form must reproduce the above copyright\n"
"   notice, this list of conditions and the following disclaimer in the\n"
"   documentation and/or other materials provided with the distribution.\n"
"\n"
"THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS\n"
"\042AS IS\042 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT\n"
"LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR\n"
"A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT\n"
"HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,\n"
"SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED\n"
"TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR\n"
"PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF\n"
"LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING\n"
"NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS\n"
"SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
  );
}

int main(int argc,char **argv)
{
  char *dev,*ele;
  int K,y;
  XGCValues gcv;
  int quit= 0;
  
  openlog("XalsaVol", 0,LOG_USER);

  dpy= XOpenDisplay(NULL);
  if (!dpy)
  {
    syslog(LOG_WARNING, "can't open display\n");
    return 1;
  }

  dev= NULL; ele= NULL;

  bgcolor= color("#333333", "black");
  ficolor= color("#888888", "white");
  licolor= color("#ffffff", "white");
  bocolor= color("#000000", "black");
  btcolor= color("#333333", "black");
  bfcolor= color("#ededed","white");

  while((K=getopt(argc,argv,"c:e:db:f:l:r:z:x:y:w:h:t:u:v"))>0)
  switch(K)
  {
  case 'c': dev= optarg; break;
  case 'e': ele= optarg; break;
  case 'd': dbg= 1; break;
  case 'b': bgcolor= color(optarg,"black"); break;
  case 'f': ficolor= color(optarg,"white"); break;
  case 'l': licolor= color(optarg,"white"); break;
  case 'r': bocolor= color(optarg,"black"); break;
  case 't': btcolor= color(optarg,"black"); break;
  case 'u': bfcolor= color(optarg,"white"); break;
  case 'z': bowidth= strtol(optarg,NULL,0); break;
  case 'x': x_pos= strtol(optarg,NULL,0); break;
  case 'y': y_pos= strtol(optarg,NULL,0); break;
  case 'w': width= strtol(optarg,NULL,0); break;
  case 'h': height= strtol(optarg,NULL,0); break;
  case '?': usage(argv[0]); break;
  case 'v': print_version(); return 0;
  }

  init_sound(dev,ele);
  if (!sound_ok) 
  {
    XCloseDisplay(dpy);
    return 1;
  }
 
  window= XCreateSimpleWindow(dpy, DefaultRootWindow(dpy), x_pos, y_pos,
                             width, height, bowidth, bocolor, bgcolor);
  if (!dbg)
  {
    XSetWindowAttributes xa;
    xa.override_redirect= True;
    XChangeWindowAttributes(dpy, window, CWOverrideRedirect, &xa);
  }
  XSelectInput(dpy, window, ExposureMask | ButtonPressMask | ButtonReleaseMask);
  XMapWindow(dpy, window);
  gcv.fill_style= FillSolid;
  gcv.line_width= 2;
  gc= XCreateGC(dpy, window, GCLineWidth | GCFillStyle, &gcv);

  init_coords();
  while(!quit)
  {
    XEvent event;
    XNextEvent(dpy, &event);
    switch(event.type)
    {
    case Expose:
       draw_window(event.xexpose.x,event.xexpose.y,
                   event.xexpose.width,event.xexpose.height);
       break;
    case ButtonPress:
       if (event.xbutton.button!=1) continue;
       y= event.xbutton.y;
       if (y<=tri_pt[0].y-3)
       {
          cros_pressed= 1;
          draw_window(0,0, width, CROSS_LENGTH*1.4);
          break;
       }
       if (y<min_y) y= min_y;
       else if (y>max_y) y= max_y;
       set_volume( (double) (max_y-y)/delta_y* (max_vol-min_vol+1) + min_vol);
       btn_pressed= 1;
       draw_window(0, rec_pt[0].y, width, 10);
       break;
    case ButtonRelease:
       if (event.xbutton.button!=1) continue;
       if (cros_pressed)
       {
          quit= 1;
          break;
       }
       btn_pressed= 0;
       if (!get_volume())
       {
         init_rec_coords();
         draw_window(0,0,width,height);
       }
       break;   
    }
  }

  if (sid) snd_mixer_selem_id_free(sid);
  if (handle) snd_mixer_close(handle);
  XCloseDisplay(dpy);
  return 0;
}
